/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  arp.c
 * Purpose: ARP handling
 * Entries: ip_resolve, arp_recv
 *
 **************************************************************************
 *
 * Copyright (C) 1995-2003 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: arp.c,v 1.5 2003/01/25 23:29:41 gkminix Exp $
 */


#include <general.h>
#include <kernel/net.h>
#include <kernel/romlib.h>
#include <pxe/common.h>
#include <pxe/undi.h>
#include "netpriv.h"
#include "arp.h"



/*
 **************************************************************************
 * 
 * Global variables
 */
static struct arphdr arpbuf;			/* ARP send buffer	*/
static struct arp_cache arp_list[ARPNUM];	/* ARP address cache	*/
static int next_arp;				/* next table entry	*/

static unsigned char bcasthw[ETH_ALEN] = {	/* broadcast address	*/
		0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

static unsigned int sizes[] = { sizeof(arpbuf) };
static unsigned char *bufs[] = { (unsigned char *)&arpbuf , NULL };



/*
 **************************************************************************
 * 
 * Find an IP address in the ARP cache. Return -1 if not found.
 */
static int findip(ip)
t_ipaddr ip;
{
  register int i;

  for (i = 0; i < ARPNUM; i++)
	if (arp_list[i].ipaddr == ip)
		break;
  return(i);
}



/*
 **************************************************************************
 * 
 * Add a new entry to the ARP cache.
 */
static void addcache(ha, ip)
unsigned char *ha;
t_ipaddr       ip;
{
  register int i;

  /*
   * If the IP address already exists in the ARP table then just replace
   * it's hardware address with the new onbe. Otherwise create a new ARP
   * table entry.
   */
  if ((i = findip(ip)) >= ARPNUM) {
	i = next_arp++;
	if (next_arp >= ARPNUM)
		next_arp = 0;
	arp_list[i].ipaddr = ip;
  }
  memcpy(arp_list[i].hwaddr, ha, ETH_ALEN);
}



/*
 **************************************************************************
 * 
 * ARP receiver routine
 */
void arp_recv(bufsize)
unsigned int bufsize;
{
#define arpreply ((struct arphdr *)(readbuf.buf))

#ifdef DEBUG
  int ret;
#endif

  /*
   * Do some checks on the received packet. It must be an ARP packet,
   * with an ethernet hardware address, and for the IP protocol. Other
   * packets just get discarded. Then save the sender's address.
   */
  if (bufsize < (sizeof(struct arphdr) + sizeof(readbuf.protocol)) ||
      readbuf.protocol != UNDI_PROT_ARP ||
      arpreply->ar_op  != htons(ARPOP_REQUEST) ||
      arpreply->ar_tip != myipaddr ||
      arpreply->ar_hrd != htons(mytype) ||
      arpreply->ar_hln != ETH_ALEN ||
      arpreply->ar_pro != htons(ARPPROT_IP) ||
      arpreply->ar_pln != IP_ALEN)
	goto recv_ret;

  arpbuf.ar_hrd = htons(mytype);
  arpbuf.ar_pro = htons(ARPPROT_IP);
  arpbuf.ar_hln = ETH_ALEN;
  arpbuf.ar_pln = IP_ALEN;
  arpbuf.ar_op  = htons(ARPOP_REPLY);
  arpbuf.ar_tip = arpreply->ar_sip;
  arpbuf.ar_sip = myipaddr;
  memcpy(arpbuf.ar_tha, arpreply->ar_sha, ETH_ALEN);
  memcpy(arpbuf.ar_sha, myhwaddr, ETH_ALEN);
#ifdef DEBUG
  ret = undi_send(arpreply->ar_sha, UNDI_PROT_ARP, 1, sizes, bufs);
  printf("ARP: reply sent, result = %x\n", ret);
#else
  undi_send(arpreply->ar_sha, UNDI_PROT_ARP, 1, sizes, bufs);
#endif

recv_ret:
  return;

#undef arpreply
}



/*
 **************************************************************************
 * 
 * Resolve IP address and return pointer to hardware address.
 */
int ip_resolve(myip, ip, gw, hwaddr)
t_ipaddr myip;
t_ipaddr ip;
t_ipaddr gw;
unsigned char **hwaddr;
{
#define arpreply ((struct arphdr *)(readbuf.buf))

  int i, counts, tout, ret;

#ifdef DEBUG
  printf("ARP: Resolving IP address %lx\n", ip);
#endif

  /* Check for default addresses */
  if (myip == n_IP_ANY)
	myip = myipaddr;
  if (gw == n_IP_ANY)
	gw = mygateway;

  /* Check that the gateway address is correct */
  if (gw != n_IP_ANY &&
      (gw == n_IP_BROADCAST || gw == myip || ((gw ^ myip) & mynetmask)))
	gw = n_IP_ANY;

  /* Check for invalid IP addresses */
  ret = PXENV_STATUS_FAILURE;
  if (ip == n_IP_ANY || ip == myip || ip == n_IP_LOCALHOST)
	goto res_ret;

  /*
   * If the destination is not on subnet, send the packets through gateway.
   * Also check for a local broadcast address.
   */
  if (ip != n_IP_BROADCAST) {
	/*
	 * This double-if construction lets the compiler produce better
	 * optimized code compared to a &&.
	 */
	if ((ip ^ myip) & mynetmask) {
		if (gw != n_IP_ANY)
			ip = gw;
		else
			goto res_ret;
	}
  }

  /* Check for broadcast address */
  ret = PXENV_STATUS_SUCCESS;
  if (ip == n_IP_BROADCAST || !((ip ^ n_IP_BROADCAST) & ~mynetmask)) {
	memcpy(hwaddr, bcasthw, ETH_ALEN);
	goto res_ret;
  }

  /* Check if we got the address in the cache already */
  if ((i = findip(ip)) < ARPNUM) {
	memcpy(hwaddr, arp_list[i].hwaddr, ETH_ALEN);
	goto res_ret;
  }

  /* No, we have to broadcast a request */
  arpbuf.ar_hrd = htons(mytype);
  arpbuf.ar_pro = htons(ARPPROT_IP);
  arpbuf.ar_hln = ETH_ALEN;
  arpbuf.ar_pln = IP_ALEN;
  arpbuf.ar_op  = htons(ARPOP_REQUEST);
  arpbuf.ar_sip = myip;
  arpbuf.ar_tip = ip;
  memcpy(arpbuf.ar_sha, myhwaddr, ETH_ALEN);
  memset(arpbuf.ar_tha, 0, ETH_ALEN);
  bufs[1] = getds();

  /*
   * Send the ARP request and wait for an answer. This is a double loop
   * which retries when no answer has been received in a timeout time.
   */
  tout = ARP_TIMEOUT;
  for (counts = 0; counts < ARP_COUNTS; counts++) {
	/* Send the request */
	if ((ret = undi_send(NULL, UNDI_PROT_ARP, 1, sizes, bufs))
						 != PXENV_STATUS_SUCCESS)
		goto res_ret;

	/* Wait for someone to answer the broadcast request */
	set_timeout(tout);
	while (!chk_timeout()) {
		/*
		 * Read next packet from UNDI and check that it's a
		 * correct ARP reply packet.
		 */
		i = undi_recv(sizeof(readbuf), (unsigned char *)&readbuf);
		if (i < (sizeof(struct arphdr) + sizeof(readbuf.protocol)) ||
		    readbuf.protocol != UNDI_PROT_ARP ||
		    arpreply->ar_op  != htons(ARPOP_REPLY) ||
		    arpreply->ar_hrd != htons(mytype) ||
		    arpreply->ar_hln != ETH_ALEN ||
		    arpreply->ar_pro != htons(ARPPROT_IP) ||
		    arpreply->ar_pln != IP_ALEN)
			continue;

		/* Add new IP address to cache and set return value */
		addcache(arpreply->ar_sha, arpreply->ar_sip);
		memcpy(hwaddr, arpreply->ar_sha, ETH_ALEN);
		goto res_ret;
	}

	/* Double timeout value for next round */
	tout <<= 1;
  }
  ret = PXENV_STATUS_ARP_TIMEOUT;

res_ret:
#ifdef DEBUG
  printf("ARP: result = %x\n", ret);
#endif
  return(ret);

#undef arpreply
}

