/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  udp.c
 * Purpose: UDP and IP handling
 * Entries: udp_open, udp_close, udp_read, udp_write
 *
 **************************************************************************
 *
 * Copyright (C) 1995-2003 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: udp.c,v 1.6 2003/01/25 23:29:41 gkminix Exp $
 */


#include <general.h>
#include <kernel/net.h>
#include <kernel/arpa.h>
#include <kernel/romlib.h>
#include <pxe/common.h>
#include <pxe/undi.h>
#include <pxe/udp.h>
#include "netpriv.h"
#include "arp.h"
#include "ip.h"
#include "udp.h"


/*
 **************************************************************************
 * 
 * Public variables:
 *
 * Note that these are all (except mymtu, readbuf and udpoflag) in network
 * order!!!
 */
struct   iobuf    readbuf;			/* global read buffer	*/
unsigned char     myhwaddr[ETH_ALEN];		/* my own hardware addr	*/
unsigned int      mymtu;			/* my own MTU		*/
unsigned int      mytype;			/* my type of network	*/
         t_ipaddr myipaddr;			/* my own IP address	*/
         t_ipaddr mynetmask;			/* my own netmask	*/
         t_ipaddr mygateway;			/* my gateway to use	*/
         int      udpoflag;			/* TRUE if open called	*/



/*
 **************************************************************************
 * 
 * Global variables:
 *
 * These are also all in network order!!!
 */
static unsigned short usource;		/* Source port number		*/
static unsigned short udest;		/* Destination port number	*/
static t_ipaddr       udaddr;		/* Destination IP address	*/
static t_ipaddr       usaddr;		/* Source IP address		*/
static t_ipaddr       gateway;		/* Gateway IP address		*/



/*
 **************************************************************************
 * 
 * Open a new UDP socket.
 */
int udp_open(daddr, gw, source, dest)
t_ipaddr daddr;
t_ipaddr gw;
int      source;
int      dest;
{
  /* Set global variables */
  usource = source;
  udest = dest;
  usaddr = myipaddr;
  udaddr = daddr;
  gateway = gw;

  /* Open UNDI interface */
  return(undi_open());
}



/*
 **************************************************************************
 * 
 * Close a UDP socket
 */
int udp_close()
{
  return(undi_close());
}



/*
 **************************************************************************
 * 
 * Receive a UDP packet. If the packet is fragmented, read the next
 * fragment and add it to the end of the buffer. This routine is non-
 * blocking.
 */
static unsigned int udp_recv(buf, bufseg, bufsize)
unsigned char *buf;
unsigned int   bufseg;
unsigned int   bufsize;
{
#define ipp ((struct iphdr *)readbuf.buf)
#define udpp ((struct udphdr *)(readbuf.buf + IP_MIN_HSIZE))
#define datap ((unsigned char *)(readbuf.buf + IP_MIN_HSIZE + sizeof(struct udphdr)))

  int fragflag;
  unsigned int fragoff;
  unsigned int readlen;
  unsigned int size;
  static t_ipaddr dest;
  static struct pshdr psbuf;
  static struct udphdr udpbuf;
  static unsigned int totlen;
  static unsigned int pktid = 0xFFFF;

  /* Loop as long as we have packets to receive */
  while ((readlen = undi_recv(sizeof(readbuf), (unsigned char *)&readbuf)) >
						sizeof(readbuf.protocol)) {

	/* If this is an ARP packet, do handle it */
	if (readbuf.protocol == UNDI_PROT_ARP) {
		arp_recv(readlen);
		continue;
	}

	/*
	 * Do some checks on the received packet. It must be an IP packet
	 * with a valid version number and header size. Also the checksum
	 * of the header must be correct.
	 */
	size = ntohs(ipp->ip_tot_len) - IP_MIN_HSIZE;
	if (
	    /* It has to be an IP packet */
	    readbuf.protocol != UNDI_PROT_IP ||

	    /* The packet must at least contain an IP header */
	    readlen <= IP_MIN_HSIZE ||

	    /* The size value in the packet has to be correct */
	    size > (sizeof(readbuf.buf) - IP_MIN_HSIZE) ||

	    /* Check for valid version number without options */
	    ipp->ip_ihl_version != ((IP_VERSION << 4) | (IP_MIN_HSIZE >> 2)) ||

	    /* Check for correct checksum */
	    ipchksum((unsigned char *)ipp, IP_MIN_HSIZE) != 0 ||

	    /* The packet has to be for UDP */
	    ipp->ip_protocol != IPPROTO_UDP)
		continue;

#ifdef DEBUG
	printf("UDP: got packet from %lx.%d for local port %d\n",
			ntohl(ipp->ip_daddr), ntohs(udpp->udp_source),
			ntohs(udpp->udp_dest));
#endif

	/*
	 * Check if this packet is fragmented and there are more frag-
	 * ments following.
	 */
	fragflag = ntohs(ipp->ip_frag_off) & IP_FRAG_MORE;
	fragoff = (ntohs(ipp->ip_frag_off) & IP_FRAG_MASK) << 3;

	/*
	 * If there are no fragments at all or this is the first fragment,
	 * we have to analyze the UDP header and then copy the data contents
	 * into the destination buffer.
	 */
	if (fragoff == 0) {
		/* Check that we have at least a UDP header */
		if (size < sizeof(struct udphdr))
			continue;

		/* Compute total size of data in packet */
		pktid = 0xFFFF;
		totlen = ntohs(udpp->udp_len) - sizeof(struct udphdr);
		if (totlen > bufsize)
			break;

		/* Save UDP header and pseudo header for later */
		psbuf.srcip = ipp->ip_saddr;
		psbuf.destip = ipp->ip_daddr;
		psbuf.pad = 0;
		psbuf.protocol = ipp->ip_protocol;
		psbuf.length = udpp->udp_len;
		memcpy((unsigned char *)&udpbuf, (unsigned char *)udpp,
								sizeof(udpbuf));

		/*
		 * Check that this packet is for our address and port
		 */
		if ((usaddr != n_IP_ANY && usaddr != ipp->ip_daddr) ||
		    (usource != 0 && usource != udpp->udp_dest))
			continue;

		/* Save packet ID and destination address for later */
		pktid = ipp->ip_id;
		dest = ipp->ip_daddr;
	}

	/*
	 * If this is a fragment, we have to copy the fragment data at
	 * the correct place within the destination buffer.
	 */
	if (fragflag || fragoff > 0) {
		/*
		 * Check that this fragment is for the same packet by
		 * using the packet ID and destination address. If this
		 * packet is invalid, discard it.
		 */
		if (ipp->ip_id != pktid ||
		    ipp->ip_daddr != dest)
			continue;

		/* Copy the fragment into the destination buffer */
		if (fragoff == 0) {
			/* First fragment */
			size -= sizeof(struct udphdr);
			fmemcpy((unsigned int)buf, bufseg, datap, size);
		} else {
			/* Middle or last fragment */
			fragoff -= sizeof(struct udphdr);
			if ((fragoff + size) > bufsize)
				continue;
			fmemcpy(((unsigned int)buf + fragoff), bufseg,
							datap, size);
		}

		/*
		 * Check if this was the last fragment. If not we retry
		 * to read a new packet. Since this routine is non-blocking,
		 * this will likely return us to the caller with zero bytes
		 * read. Therefore, the caller has to repeat this call until
		 * all fragments have been read. In that case, we return the
		 * actual number of bytes read.
		 */
		if (fragflag)
			/* Wait for next fragment */
			continue;
		else
			/* Last fragment, check for checksum */
			goto recv_end;
	}

	/*
	 * If we turn up here the packet is not fragmented. We can simply
	 * copy it into the destination buffer. The packet size can't be
	 * larger than the buffer size as that has already been checked
	 * above.
	 */
	fmemcpy((unsigned int)buf, bufseg, datap, totlen);

recv_end:
	/*
	 * Perform a checksum run to see if the packet has been received
	 * correctly. This has only to be done when the checksum in the
	 * UDP packet is not FFFF.
	 */
	if (udpbuf.udp_check && udpbuf.udp_check != 0xFFFF)
		if (udpchksum((unsigned char *)&psbuf, sizeof(psbuf),
		    (unsigned char *)&udpbuf, sizeof(udpbuf),
		    (unsigned char *)buf, bufseg, totlen) != 0)
			break;

	/* Setup some global variables. This is required for TFTP and DHCP */
	udest = udpbuf.udp_source;
	udaddr = psbuf.srcip;

	/* Reset packet ID and return with number of bytes read */
	pktid = 0xFFFF;
	return(totlen);
  }
  return(0);

#undef ipp
#undef udpp
#undef datap
}



/*
 **************************************************************************
 * 
 * Read one packet from a UDP socket. This routine is blocking.
 */
int udp_read(buf, bufseg, bufsize, timeout, abortch)
unsigned char *buf;
unsigned int   bufseg;
unsigned int  *bufsize;
int            timeout;
char           abortch;
{
  unsigned int buflen;

  /* Wait until we receive a correct packet */
  set_timeout(timeout);
  while (!chk_timeout()) {
	if (abortch && chkkey() == abortch)
		return(PXENV_STATUS_ARP_CANCELED);
	if ((buflen = udp_recv(buf, bufseg, *bufsize)) > 0) {
		*bufsize = buflen;
		return(PXENV_STATUS_SUCCESS);
	}
  }
  return(PXENV_STATUS_ARP_TIMEOUT);
}



/*
 **************************************************************************
 * 
 * Write one packet to a UDP socket
 */
int udp_write(buf, bufseg, len)
unsigned char *buf;
unsigned int   bufseg;
unsigned int   len;
{
  int status;
  unsigned char udhw[ETH_ALEN];
  unsigned int sizes[3];
  unsigned char *bufs[6];
  static unsigned int packetid;
  static struct pshdr psbuf;
  static struct udphdr udpbuf;
  static struct iphdr ipbuf;

  /* Determine the hardware address of the recipient */
  if ((status = ip_resolve(usaddr, udaddr, gateway, &udhw)) !=
						PXENV_STATUS_SUCCESS)
	goto write_ret;

#ifdef DEBUG
  printf("UDP: sending %d bytes from %ld.%d to %ld.%d (%x:%x:%x:%x:%x:%x)\n",
			len,
			ntohl(usaddr), ntohs(usource),
			ntohl(udaddr), ntohs(udest),
			udhw[0], udhw[1], udhw[2], udhw[3], udhw[4], udhw[5]);
#endif

  /* Prepare UDP datagram */
  udpbuf.udp_source = usource;
  udpbuf.udp_dest = udest;
  udpbuf.udp_len = htons(len + sizeof(udpbuf));
  udpbuf.udp_check = 0;

  /* Prepare UDP pseudo header for calculating the checksum */
  psbuf.srcip = usaddr;
  psbuf.destip = udaddr;
  psbuf.pad = 0;
  psbuf.protocol = IPPROTO_UDP;
  psbuf.length = udpbuf.udp_len;

  /* Compute UDP checksum */
  udpbuf.udp_check = udpchksum((unsigned char *)&psbuf, sizeof(psbuf),
				(unsigned char *)&udpbuf, sizeof(udpbuf),
				(unsigned char *)buf, bufseg, len);

  /* Prepare IP header */
  ipbuf.ip_ihl_version = ((IP_VERSION << 4) | ((sizeof(ipbuf)/4) & 0x0f)) & 0xff;
  ipbuf.ip_tos         = IP_STD_TOS;
  ipbuf.ip_tot_len     = htons(sizeof(ipbuf) + sizeof(udpbuf) + len);
  ipbuf.ip_id          = htons(++packetid);
  ipbuf.ip_frag_off    = 0;
  ipbuf.ip_check       = 0;
  ipbuf.ip_ttl         = IP_TTL;
  ipbuf.ip_protocol    = IPPROTO_UDP;
  ipbuf.ip_daddr       = udaddr;
  ipbuf.ip_saddr       = usaddr;
  ipbuf.ip_check       = ipchksum((unsigned char *)&ipbuf, sizeof(ipbuf));

  /* Finally send out the packet */
  if (len > (mymtu - sizeof(struct iphdr) - sizeof(struct udphdr))) {
	/* We need to fragment the outgoing packet */
	unsigned int fraglen;
	unsigned int fragoff;

	/* Broadcast packets can't get fragmented */
	if (udaddr == n_IP_BROADCAST ||
	    !((udaddr ^ n_IP_BROADCAST) & ~mynetmask)) {
		status = PXENV_STATUS_FAILURE;
		goto write_ret;
	}

	/* Compute length of each fragment. Has to be on 8 byte boundary */
	fraglen = (mymtu - sizeof(ipbuf)) & ~0x0007;

	/* Setup first packet */
	ipbuf.ip_frag_off = htons(IP_FRAG_MORE);
	ipbuf.ip_tot_len = htons(sizeof(ipbuf) + fraglen);
	ipbuf.ip_check = 0;
	ipbuf.ip_check = ipchksum((unsigned char *)&ipbuf, sizeof(ipbuf));

	/* Write first packet */
	sizes[0] = sizeof(ipbuf);
	sizes[1] = sizeof(udpbuf);
	sizes[2] = fraglen - sizeof(udpbuf);
	bufs[0] = (unsigned char *)&ipbuf;
	bufs[2] = (unsigned char *)&udpbuf;
	bufs[1] = bufs[3] = (unsigned char *)getds();
	bufs[4] = buf;
	bufs[5] = (unsigned char *)bufseg;
	if ((status = undi_send(udhw, UNDI_PROT_IP, 3, sizes, bufs)) !=
							PXENV_STATUS_SUCCESS)
		goto write_ret;
	
	/* Now write all full fragments */
	fragoff = fraglen;
	buf += fraglen - sizeof(udpbuf);
	len -= fraglen - sizeof(udpbuf);
	while (len > fraglen) {
		/* Setup packet for next fragment */
		ipbuf.ip_frag_off = htons(IP_FRAG_MORE | (fragoff >> 3));
		ipbuf.ip_tot_len = htons(sizeof(ipbuf) + fraglen);
		ipbuf.ip_check = 0;
		ipbuf.ip_check = ipchksum((unsigned char *)&ipbuf, sizeof(ipbuf));

		/* Write next packet */
		sizes[1] = fraglen;
		bufs[2] = buf;
		bufs[3] = (unsigned char *)bufseg;
		if ((status = undi_send(udhw, UNDI_PROT_IP, 2, sizes, bufs)) !=
							PXENV_STATUS_SUCCESS)
			goto write_ret;

		/* Proceed with next packet */
		fragoff += fraglen;
		buf += fraglen;
		len -= fraglen;
	}	

	/* Setup packet for last fragment */
	ipbuf.ip_frag_off = htons(fragoff >> 3);
	ipbuf.ip_tot_len = htons(sizeof(ipbuf) + len);
	ipbuf.ip_check = 0;
	ipbuf.ip_check = ipchksum((unsigned char *)&ipbuf, sizeof(ipbuf));

	/* Write last packet */
	sizes[1] = len;
	bufs[2] = buf;
	bufs[3] = (unsigned char *)bufseg;
	status = undi_send(udhw, UNDI_PROT_IP, 2, sizes, bufs);
  } else {
	/* No fragmentation necessary */
	sizes[0] = sizeof(ipbuf);
	sizes[1] = sizeof(udpbuf);
	sizes[2] = len;
	bufs[0] = (unsigned char *)&ipbuf;
	bufs[2] = (unsigned char *)&udpbuf;
	bufs[1] = bufs[3] = (unsigned char *)getds();
	bufs[4] = buf;
	bufs[5] = (unsigned char *)bufseg;
	status = undi_send(udhw, UNDI_PROT_IP, 3, sizes, bufs);
  }

write_ret:
#ifdef DEBUG
  printf("UDP: send status %d\n", status);
#endif
  return(status);
}



/*
 **************************************************************************
 * 
 * UDP open PXE API function
 */
static int pxe_udp_open(params)
t_udp_open *params;
{
  int status;

  /* Check that we are allowed to open a UDP channel */
  if (udpoflag)
	return(PXENV_STATUS_UDP_OPEN);
  if (tftpoflag)
	return(PXENV_STATUS_TFTP_OPEN);

  /* Set my own IP address */
  if (params->src_ip == n_IP_ANY)
	params->src_ip = myipaddr;
  usaddr = params->src_ip;

  /* Finally open the network interface */
  if ((status = undi_open()) == PXENV_STATUS_SUCCESS)
	udpoflag = TRUE;
  return(status);
}



/*
 **************************************************************************
 * 
 * UDP close PXE API function
 */
static int pxe_udp_close(params)
t_udp_close *params;
{
  int status;

  /* Check that we are allowed to close the UDP channel */
  if (!udpoflag)
	return(PXENV_STATUS_UDP_CLOSED);

  /* Close the network interface */
  if ((status = undi_close()) == PXENV_STATUS_SUCCESS)
	udpoflag = FALSE;
  return(status);
}



/*
 **************************************************************************
 * 
 * UDP read PXE API function
 */
static int pxe_udp_read(params)
t_udp_read *params;
{
  int status;
  unsigned int buflen = 0;
  t_ipaddr oldsaddr = usaddr;

  /* Check that we are allowed to rerad from the UDP channel */
  if (!udpoflag)
	return(PXENV_STATUS_UDP_CLOSED);

  /* Set global variables for this call only */
  usource = params->d_port;
  udest = htons(-1);
  usaddr = params->dest_ip;
  udaddr = n_IP_ANY;

  /* Read a packet using UNDI */
  buflen = udp_recv(params->bufofs, params->bufseg, params->bufsize);
  if (buflen == 0)
	status = PXENV_STATUS_FAILURE;
  else {
	params->src_ip = udaddr;
	params->s_port = udest;
	params->bufsize = buflen;
	status = PXENV_STATUS_SUCCESS;
  }
  usaddr = oldsaddr;
  return(status);
}



/*
 **************************************************************************
 * 
 * UDP write PXE API function
 */
static int pxe_udp_write(params)
t_udp_write *params;
{
  /* Check that we are allowed to write to the UDP channel */
  if (!udpoflag)
	return(PXENV_STATUS_UDP_CLOSED);

  /* Set global variables and write the packet */
  if (params->src_port == 0)
	usource = ntohs(PXE_UDP_DEF_PORT);
  else
	usource = params->src_port;
  udest = params->dest_port;
  udaddr = params->ip;
  gateway = params->gw;
  return(udp_write(params->bufofs, params->bufseg, params->bufsize));
}



/*
 **************************************************************************
 * 
 * Function table for UDP PXE functions
 */
struct functab udp_func_tab = {
	UDP_MINNUM, UDP_MAXNUM,
	{{ pxe_udp_open,	sizeof(t_udp_open)},
	 { pxe_udp_close, 	sizeof(t_udp_close)},
	 { pxe_udp_read, 	sizeof(t_udp_read)},
	 { pxe_udp_write, 	sizeof(t_udp_write)}
	}
};

